# -*- coding: utf-8 -*-
"""
Created on Wed Dec  7 16:13:12 2016

@author: akidribm
"""

# Imports sortal sort types, functions and rules
from sortal import (transformSet, similarityMap, rule, primitiveSort, point2D, coordinate, lineSegment2D, line2D, label, weight, disjunctiveSort, attributeSort, affineTransform,  vector2D)
from rule3_moveSquareAndShape import rule3a, rule3b
from rule2_90degreeTriangle import rule2

# Imports API functions and variables
from api_create import create
from form_algebra import form_algebra
from sort_types import sort_types

# Imports GUI Library
from pylab import *
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import  matplotlib.pylab as plt
import matplotlib.pyplot as pltplot
from matplotlib.collections import LineCollection
matplotlib.use ("gTkAgg")
import tkinter as tk
from tkinter import ttk, Tk, Label, Button, Frame
from tkinter import *

plt.ion()

"""
===========
API METHODS
===========
============
FORM ALGEBRA
============
'Form Algebra' methods do not change the form that serves as 
the calling function but return a new instance after the form algebra as been performed

METHODS:
  
SUM where new_form_1 is the union of formL1 and formL3
    new_form_1 = algebra.sum(formL1, formL3)

DIFFERENCE where new_form_2 is the form containing the set of elements left 
           after those in common between formL1 and formL3 are removed
           new_form_2 = algebra.difference(formL1, formL3)

PRODUCT where new_form_3 is the intersection of formL1 and formL3
        new_form_3 = algebra.product(formL1, formL3)

=============
FORM CREATION
=============
'Create' methods generate and return forms based on input. As part of the API, they are imported by 
importing create from the api_create file, i.e. from api_create import create.

All sort types are initialized and stored in the file sort_types.py, inside the dictionary sort_types.
This is imported via 'from sort_types import sort_types'

Vectors are created by calling either vector2D or vector3D, as follows:
v1 = vector2D(x,y) where (x,y) are coordinates
v2 = vector3D(x,y,z) where (x,y,z) are coordinates

METHODS:
  
Create.createForm where a form of individuals is created from a list of individuals

                  EXAMPLE: a form of line segments where ls1 = lineSegment2D(sort_types['lineSeg2D'], start, end) and so on
                  new_form1_list = [ls1, ls2, ls3, ls4, ls5, ls6, ls7, ls8, ls9, ls10] 
                  new_form1 = create.createForm(sort_types['lineSeg2D'], new_form1_list) 
                   
                  OR a form of points where p1 = point2D(sort_types['pointSort2D'], x, y) and so on
                   
                  new_form1_list = [p1, p2, p3, p4, p5, p6, p7, p8, p9, p10] 
                  new_form1 = create.createForm(sort_types['lineSeg2D'], new_form1_list)

Create.lineSegForm where a form of line segments is created from an input of vectors as lists within a list
                   Each vector pair corresponds to a line segment, where index 0 of the pair is the start
                   of the line segment and index 1 of the pair is the end of the line segment
                  
                   This method will raise an error if the sort type passed is not based on 
                   lineSegment2D or lineSegment3D, i.e. sort_types['lineSeg2D'] OR sort_types['lineSeg3D']
                     
                   new_form2_set = [[vb1, vb2], [vb2, vb3], [vb3, vb4], [vb4, vb1]]
                   new_form2_set = create.lineSegForm(sort_types['lineSeg2D'], form2_set)
                  
                   OR if with 3D vectors (vector3D), 
                  
                   new_form2_set = create.lineSegForm(sort_types['lineSeg3D'], form2_set)
         
Create.polygonForm where a form containing a polygon is created from an input of a list of vectors
                   These vectors serve as vertices and need to be listed in the order they will be connected
                   It is assumed that the last vector will be connected to the first vector
                   In the example, v10 will be connected to v1 to close the polygon
                    
                   This method will raise an error if the sort type passed is not based on 
                   lineSegment2D or lineSegment3D, i.e. sort_types['lineSeg2D'] OR sort_types['lineSeg3D']
                    
                   new_form3_set = [v1, v2, v3, v4, v5, v6, v7, v8, v9, v10]
                   new_form3 = create.polygonForm(sort_types['lineSeg2D'], new_form3_set)
                    
                   OR if with 3D vectors (vector3D), 
                  
                   new_form2_set = create.polygonForm(sort_types['lineSeg3D'], new_form3_set)
"""    
if __name__ == "__main__":
    
    LARGE_FONT= ("Verdana", 16)

    # Initializes classes
    algebra = form_algebra() # To perform form algebra and store new instance in variable
    create = create()   # Methods for creating forms
    
    # Examples of using create class methods
    """
    ================
    POLYGON CREATION
    ================
    """
    # To create 2D vectors, use vector2D(x, y) where (x, y) are vector coordinates
    v1 = vector2D (-2,0) 
    v2 = vector2D (-2,1) 
    v3 = vector2D (-1,1) 
    v4 = vector2D (-1,2) 
    v5 = vector2D (0,2) 
    v6 = vector2D (0,-2) 
    v7 = vector2D (1,-2) 
    v8 = vector2D (1,-1) 
    v9 = vector2D (2,-1) 
    v10 = vector2D (2,0) 
    
    # Creates a form with a 'polygon' inside using vectors
    # Follows the process: vectors as vertices -> form with a 'polygon' inside
    # Takes input sort type 'lineSeg2D' which is called from dictionary 'sort_types'
    formL1_set = [v1, v2, v3, v4, v5, v6, v7, v8, v9, v10]
    formL1 = create.polygonForm(sort_types['lineSeg2D'], formL1_set)

    """
    =====================================
    POLYGON WITH ADDITIONAL LINE SEGMENTS
    =====================================
    """
    va1 = vector2D (0,0) 
    va2 = vector2D (0,1) 
    va3 = vector2D (2,1) 
    va4 = vector2D (2,0) 
    
    # Creates a form with a polygon inside, then adds extra line segments in the code snippet below 
    formL2_set = [va1, va2, va3, va4]
    formL2 = create.polygonForm(sort_types['lineSeg2D'], formL2_set)

    # The vectors that correspond to the line segments' starts and ends must first be turned into line segments
    ls1 = lineSegment2D(sort_types['lineSeg2D'], va3, va1) # va3 - start, va1 - end of line segment
    ls2 = lineSegment2D(sort_types['lineSeg2D'], va4, va2) # va4 - start, va2 - end of line segment
    # ls1 and ls2 can then be inserted manually into the form
    formL2.insert(ls1)
    formL2.insert(ls2)
    
    """
    =====================
    FORM OF LINE SEGMENTS
    =====================
    """
    vb1 = vector2D (0,0) 
    vb2 = vector2D (0,2) 
    vb3 = vector2D (4,2) 
    vb4 = vector2D (4,0)
    vb5 = vector2D (0,1) 
    vb6 = vector2D (2,0) 
    vb7 = vector2D (4,1) 
    vb8 = vector2D (2,2)
    
    # Creates a form of line segments where each vector tuple [v1, v2] 
    # corresponds respectively to the start and end of a line segment
    formL3_set = [[vb1, vb2], [vb2, vb3], [vb3, vb4], [vb4, vb1], [vb3, vb1], [vb4, vb2], [vb5, vb7], [vb6, vb8], [vb5, vb6], [vb6, vb7], [vb7, vb8], [vb8, vb5]]
    formL3 = create.lineSegForm(sort_types['lineSeg2D'], formL3_set)

    # Initializes rules
    rule_1 = rule2()
    rule_2 = rule3a()
    rule_3 = rule3b()
    rules = [rule_1, rule_2, rule_3]
        
    class Application(Frame):
        print ("<<<< GENERATING APPLICATION WINDOW >>>>")
        fig_main = Figure(figsize = (3, 3), dpi = 150) # Figure contains 'Current Drawing'
        fig_options = Figure(figsize = (8, 8), dpi = 90) # Figure contains 'Matches, Results'
        fig_rules = Figure(figsize = (2, 2), dpi = 80) # Figure contains 'Rules'
 
        formL4 = algebra.sum(formL1, formL3)  # Multiplies formL1 and formL3
        form_set = [formL1, formL2, formL3, formL4] # List of available forms

        # Sets Current Drawing to formL4
        # Current Drawing can be changed by changing index of form_set
        form_test = form_set[3] 
  
        matchForms = [] # Holds list of match instances for all three rules
        resultForms = [] # Holds list of result instances for chosen match
        event_list = ['o'] # Holds list of events in the order they occur in, i.e. o - Starting Window, m - Matches, r - Results
        get_ax_index = [] # Holds list of figure subplot instances in fig_options
        rules_sides = [] # Holds list of LHS and RHS of each rule   

        """
        ================================
        APPLICATION'S MAIN FUNCTION LOOP
        ================================
        """
        def __init__(self, master = None):
            Frame.__init__(self, master)
            self.grid()
            self.master.title("Test Case 1: Shape Grammar Interpreter")
    
            # Generates tkinter window grid with 6 rows and 5 columns
            for r in range(6):
                self.master.rowconfigure(r, weight=1)    
            for c in range(5):
                self.master.columnconfigure(c, weight=1)

            # Holds 'Current Drawing' figure (fig_main)
            self.Frame1 = Frame(master, bg="gray")
            self.Frame1.grid(row = 0, column = 0, rowspan = 6, columnspan = 1, sticky = W+E+N+S) 
            self.frame1_text = "Current Drawing"
            self.Frame1_label = Label (master = self.Frame1, text = self.frame1_text, background = "gray", font = LARGE_FONT, anchor = "center")
            self.Frame1_label.pack()
            
            # Holds 'Matches/Results' figure (fig_options)
            self.Frame2 = Frame(master, bg='#3ECE90')
            self.Frame2.grid(row = 0, column = 1, rowspan = 6, columnspan = 3, sticky = W+E+N+S)
            self.Frame2_label = Label (master = self.Frame2, text = "", background = '#3ECE90', font = LARGE_FONT, anchor = "center")
            self.Frame2_label.pack()
            
            # Holds 'Rules' figure (fig_rules)
            # LHS and RHS figures are side by side for each rule
            self.Frame3 = Frame(master, bg='gray')
            self.Frame3.grid(row = 0, column = 4, rowspan = 6, columnspan = 2, sticky = W+E+N+S)
            self.Frame3_label = Label (master = self.Frame3, text = "Rules", background = 'gray', font = LARGE_FONT, anchor = "center")
            self.Frame3_label.pack()

            # Initalizes option_chosen variable, which takes input from user through mouse clicks
            # This serves as user input on which match to generate results for, which result to apply on the form
            self.option_chosen = IntVar()
            self.refresh_matchForms() # Generates initial matches 
            self.menu() # Adds menu 
            self.matchForms_viz() # Draws matches onto Frame 2
            self.select_figure(self.option_chosen) # Select_figure starts visualization of fig_main and fig_options
            self.connect() # Connects GUI to mouseclick and hover events 
 
        """
        ================================
        API: MATCH AND RESULT GENERATION
        ================================
        """
        # Refreshes matchForms and resultForms lists after one run of selection, i.e. o -> m -> r
        def refresh_matchForms(self):
            for i in rules:
                i.detect(self.form_test)
                
        # Generates matches based on which form is assigned to self.form_test
        # The contents of self.matchForms look like [[matches from Rule 1], [matches from Rule 2], [matches from Rule 3]]
        def matchForms_generate(self):
            self.matchForms = []
            self.refresh_matchForms()
            for i in rules:
                self.matchForms.append(i.detection.matches())

        # Generates results based on chosen match from option_chosen variable
        def resultForms_generate(self, option_chosen):
            self.resultForms = [] # Stores result forms 
            self.matchF = self.matchForms_list[option_chosen - 1] # Stores chosen match from list of match forms based on user input
            self.matchRule =  self.rules_log[option_chosen - 1] - 1 # Stores which rule was applied to generate the chosen match stored in self.matchF
            self.resultForms.append(rules[self.matchRule].detection.results(self.matchF)) # Stores result forms generated from chosen match self.matchF with corresponding rule

        # Applies rule to form_test based on chosen match and chosen result
        # rules[self.matchRule].detection.applications contains the results based on the match and the rule
        def confirmForm (self, option_chosen):
            for i in rules[self.matchRule].detection.applications:
                if i.resultForm == self.resultForms_list[option_chosen-1]: # Checks which rule result corresponds to the chosen result generated by the match forms
                    i.perform (self.form_test) # Performs rule on form based on chosen result
                    
        """
        ========================
        GRAPHICAL USER INTERFACE
        ========================
        ==========
        EVENT TYPE
        ==========
        Event types refer to the phase the GUI is in.
        
        'o' : Match selection
        
        'm' : Result selection based on match choice
        
        'r' : Result has been selected; form has been transformed/rule has been applied 
              based on result choice, continue selecting matches
              
        One selection run consists of o -> m -> r.
        If an error is triggered (e.g. when the mouse click is outside the subplot box), 
        then the current event type is appended again to event_list.
        """
        # Changes Frame 2 Label based on event type
        def text_configure(self, event_type, option_chosen):
            # If first iteration ('o') or results have been chosen ('m' preceeded by 'r'), label shows the number of matches
            if event_type == 'o' or event_type == 'r': 
                self.Frame2_label.configure(text = "Matches: {}".format(len(self.matchForms_list)))
            elif event_type == 'm': # If match has been selected, shows number of results based on match selected and corresponding rule
                self.Frame2_label.configure(text = "Results: {}".format(len(self.resultForms_list)) + ", Applying Rule {}".format(self.rules_log[option_chosen-1]))                
        
        # Changes options based on event type to Matches/Results
        def visualization(self, event_type, option_chosen):
            if event_type == 'o':
                self.viz_main(self.form_test, self.fig_main, self.Frame1, create_plot = 'y') # Draws 'Current Drawing'
                self.rules_viz() # Draws 'Rules' plots
                self.matchForms_viz() # Draws 'Matches' plots
            elif event_type == 'm':
                self.resultForms_viz(option_chosen) # Draws 'Results' plots and redraws 'Current Drawing' by superimposing chosen match
            elif event_type == 'r':
                self.confirmForm_viz(option_chosen) #  # Applies chosen result's rule to form_test and redraws 'Current Drawing'
                self.matchForms_viz() # Generates new matches based on form_test after rule has been applied and updates 'Matches/Results' frame
         
        # Calls visualization and text_configure functions to execute event based on self.option_chosen input
        # Adds next event to self.event_list queue if no error is triggered
        def select_figure(self, option_chosen): 
            try:
                event_type = self.event_list[-1]  
                self.visualization(event_type, option_chosen)
                self.text_configure(event_type, option_chosen)
                
                if event_type == 'o' or event_type == 'r':
                    self.event_list.append('m')
                elif event_type == 'm':
                    self.event_list.append('r')
            except:
                self.error_box()
                self.event_list.append(event_type)
                
        # Visualizes rules' LHS and RHS on rightmost part of GUI window
        def rules_viz(self):
            self.rules_sides = []
            for i in rules:
                self.rules_sides.append(i.lhs)
                self.rules_sides.append(i.rhs)
            self.viz_options(self.rules_sides, self.fig_rules, self.Frame3, create_plot = 'y', create_label ='ru', col_count = 2, onclick_option = 'n')

        # Initializes menu
        def menu(self):
            menubar = Menu (root)  # Saves PDF of 'Current Drawing' only
            menubar.add_command(label="Save Current Drawing", command=self.save) # Closes window
            menubar.add_command(label="Exit", command=root.destroy)
            root.config(menu=menubar)
       
        # Saves file of 'Current Drawing'
        # File type may be changed by change the .pdf extension to e.g.: .jpg, .png 
        def save(self):
            self.fig_main.savefig('currentdrawing.pdf', bbox_inches='tight')
            
        # Identifies which subplot the mouseclick was in, calls self.select_figure() to execute next event
        def on_press(self, event):
            for i, ax in enumerate(self.get_ax_index): 
               if ax == event.inaxes:
                   self.option_chosen = i+1
            self.select_figure(self.option_chosen)
        
        # Identifies which subplot the mouse is hovering over, changes bg color of chosen subplot
        def on_motion(self, event):
            for ax in self.get_ax_index: 
               if ax == event.inaxes:
                   ax.set_axis_bgcolor('0.55')  
            self.fig_options.canvas.draw_idle()

        # Reverses bg color of subplot to white after mouse cursor leaves the subplot
        def off_motion(self, event):
            for ax in self.get_ax_index: 
                   if ax == event.inaxes:
                       ax.set_axis_bgcolor("white")
            self.fig_options.canvas.draw_idle()        
        
        # Connects mouse event functions to the 'Matches/Results' figure (fig_options)
        def connect(self):
            self.cidpress = self.fig_options.canvas.mpl_connect('button_press_event', self.on_press)
            self.cidmotion_on = self.fig_options.canvas.mpl_connect('axes_enter_event', self.on_motion)
            self.cidmotion_off = self.fig_options.canvas.mpl_connect('axes_leave_event', self.off_motion)

        # Displays Error Dialog Box when the select_figure function gets an error or when the mouse click is outside the fig_options subplot areas
        def error_box(self):
            error_box = Tk()
            error_box.title("Input Error")
            if self.event_list[-1] == 'm':
                Label(error_box, text = "Please click directly on the plot of the match.", padx = 5, pady = 15).pack()
            elif self.event_list[-1] == 'r':
                Label(error_box, text = "Please click directly on the plot of the result.", padx = 5, pady = 15).pack()
            Button(error_box, text= "OK", padx = 10, pady = 5, command= error_box.destroy).pack()  
            error_box.update()
          
        # Generates and plots matchForms in 'Matches/Results' frame (Frame 2)            
        def matchForms_viz(self):
            self.rules_log = []
            self.matchForms_list = []
            self.matchForms_generate()
            self.super_forms = [] 
            
            # Turns self.matchForms into a single-layered list from the list of lists and uses 
            # the index of the list inside the list as the rule number
            # Generates copies of form_test to superimpose matchForms on
            for a, i in enumerate(self.matchForms):
                for j in i:
                    self.matchForms_list.append(j)
                    self.super_forms.append(self.form_test)
                    self.rules_log.append(a+1)

            # Updates 'Matches/Results' figure by drawing match options superimposed on form_test
            self.fig_options.clf()
            if len(self.event_list) == 1:
                self.viz_options(self.super_forms, self.fig_options, self.Frame2, create_plot = 'y', super_color = '#3ECE90')
                self.event_list.append('o')
            else:
                self.viz_options(self.super_forms, self.fig_options, self.Frame2, super_color = '#3ECE90')
            self.viz_options(self.matchForms_list, self.fig_options, self.Frame2, super_color = "blue", create_label = "m")
            self.fig_options.canvas.draw_idle()
                    
        # Generates and plots matchForms in 'Matches/Results' frame (Frame 2) 
        def resultForms_viz(self, option_chosen):
            self.resultForms_generate(option_chosen) # Generates results based on chosen match 
            self.resultForms_list = []
            super_forms = [] 
            matchForm_sup = [] 
            for i in self.resultForms:
                for j in i:
                    self.resultForms_list.append(j)
                    super_forms.append(self.form_test) # For superimposing result figures onto form_test
                    matchForm_sup.append(self.matchF) # For superimposing chosen match figures onto form _test

            # Updates 'Current Drawing' with chosen match superimposed on form_test figure
            self.fig_main.clf()
            self.viz_main(self.form_test, self.fig_main, self.Frame1)
            self.viz_main(self.matchF, self.fig_main, self.Frame1, super_color = "blue")
            self.fig_main.canvas.draw_idle()
            
            # Updates 'Matches/Results' with results generated from chosen match
            # Draws both match and result (superimposed) on form
            self.fig_options.clf()
            self.viz_options(super_forms, self.fig_options, self.Frame2, super_color = '#3ECE90')
            self.viz_options(matchForm_sup, self.fig_options, self.Frame2, super_color = "blue")
            self.viz_options(self.resultForms_list, self.fig_options, self.Frame2, super_color = '#FCA309', create_label = "r")
            self.fig_options.canvas.draw_idle()
        
        # Applies rule of chosen result to form_test and redraws 'Current Drawing' after rule has been applied
        def confirmForm_viz(self, option_chosen):
            self.confirmForm(option_chosen)
            # Updates 'Current Drawing' with form_test after rule application   
            self.fig_main.clf()
            self.viz_main(self.form_test, self.fig_main, self.Frame1)
            self.fig_main.canvas.draw_idle()
            
        # Extracts and prepares line coordinates of each line segment in form for visualization 
        # For this case, a line segment is two sets of vertices connected to each other, so plotting is 
        # from [x1, y1] to [x2, y2], from [x3, y3] to [x4, y4] and so on
        def data_prep(self, line_set, w, h):
            # w - number of coordinates, e.g. (x, y) -> w = 2
            # h - number of lines in form
            lineSeg_x = [[0 for x in range(w)] for y in range(h)]
            lineSeg_y = [[0 for x in range(w)] for y in range(h)]
            
            for i in range (h):
                for j in range (w):
                    if j == 0:  # For x,y and potentially z coordinates
                        if line_set[i].start.type == 'struct-vector2Dfloat':
                            lineSeg_x[i][j] = float(round(line_set[i].start.x, 2))
                            lineSeg_y[i][j] = float(round(line_set[i].start.y, 2))
                        elif line_set[i].start.type == 'struct-vector2D':
                            lineSeg_x[i][j] = float(line_set[i].start.x.value)
                            lineSeg_y[i][j] = float(line_set[i].start.y.value)
                    else:
                        if line_set[i].end.type == 'struct-vector2Dfloat':
                            lineSeg_x[i][j] = line_set[i].end.x
                            lineSeg_y[i][j] = line_set[i].end.y
                        elif line_set[i].start.type == 'struct-vector2D':
                            lineSeg_x[i][j] = line_set[i].end.x.value
                            lineSeg_y[i][j] = line_set[i].end.y.value
            return lineSeg_x, lineSeg_y
            
        # Extracts individual line segments from form and puts them all together in a list
        def call_form (self, form):
            result_form = [0 for x in range(len(form.elements))]
            for x in range(len(form.elements)):
                result_form[x] =  form.elements[x]
            return result_form
         
        # Updates 'Current Drawing' frame (Frame 1)
        def viz_main(self, form, figure_House, Frame_house, super_color = '#3ECE90', create_plot = 'n'):
            result_form = self.call_form (form)
            lineSeg_x, lineSeg_y = self.data_prep(result_form, 2, h = len (result_form))
            self.a_main = self.fig_main.add_subplot(1,1,1)
            
            # Formats 'Current Drawing' Figure
            self.a_main.axes.set_anchor('W')   
            self.a_main.margins(y=0.3)
            self.a_main.margins(x=0.1)
            self.a_main.axes.get_xaxis().set_visible(False)
            self.a_main.axes.get_yaxis().set_visible(False)
            self.a_main.axes.set_aspect(0.9)
            
            # Plots Current Drawing in Frame 1
            for i in range(len(result_form)):
                self.a_main.plot(lineSeg_x[i], lineSeg_y[i], color = super_color, linewidth = 1.5) 
              
            # Creates entirely new figure
            if create_plot == 'y':
                self.canvas_main = FigureCanvasTkAgg(figure_House, Frame_house)
                self.canvas_main.get_tk_widget().pack(fill = BOTH, expand = 1) 
                
        # Updates 'Matches/Results' frame (Frame 2) 
        def viz_options(self, form_list, figure_House, Frame_house, create_plot = 'n', super_color = '#3ECE90', create_label =  "n", col_count = 3, onclick_option = 'y'):
            # Refreshes index counter, list of subplot instances, and adjusts column count based on number of forms to be visualized
            self.get_ax_index = []
            if len(form_list) > 12:
                col_count = 4
             
            ind = 1 # Keeps track of subplot number
            # Plots Matches/Results in Frame 2
            for i in form_list:
                row_count = ceil(len(form_list)/col_count)
                result_form = self.call_form (i)
                lineSeg_x, lineSeg_y = self.data_prep(result_form, 2, h = len (result_form))
                self.a_options = figure_House.add_subplot(row_count, col_count, ind)

                # Adjusts ratio of height and width of each subplot in 'Matches/Results' depending on number of subplots generated.
                if len(form_list) < 20:
                    self.a_options.axes.set_aspect(1.1)
                else:
                    self.a_options.axes.set_aspect("equal")
                    
                for j in range(len(result_form)):            
                    if create_label == "m":
                        self.a_options.set_title('%d'% (ind,)+': Rule {}'.format(self.rules_log[ind-1]))
                    elif create_label == "r":
                        self.a_options.set_title('Result {}'.format(ind))
                    elif create_label == "ru":
                        self.a_options.set_xlim(-2,3.5)
                        self.a_options.set_ylim(-2.5,2.5)
                        extra_label=''
                        # Sets text inside title for 'Rules' subplots
                        if ind <=2:
                            counter = 1 
                        elif ind > 2 and ind <=4:
                            counter = 2
                        elif ind > 4 and ind <=6:
                            counter = 3
                            
                        if ind%2 == 0:
                            extra_label = "RHS\n"
                        else:
                            extra_label = "LHS\n"

                        self.a_options.set_title(extra_label + 'Rule {}'.format(counter), fontsize = 11)
                        
                    # Plots line segment in subplot
                    self.a_options.plot(lineSeg_x[j], lineSeg_y[j], color = super_color, linewidth = 1.3, picker = 2)

                self.a_options.axes.get_xaxis().set_visible(False)
                self.a_options.axes.get_yaxis().set_visible(False) 
                self.a_options.margins(y=0.2)
                self.a_options.margins(x=0.2)
                ind = ind + 1   
                # Gets instance of subplot for onclick function
                if onclick_option == 'y':
                    self.get_ax_index.append(self.a_options) 
                    
            figure_House.subplots_adjust(hspace=0.8)
            # Creates entirely new figure
            # If a figure has already been plotted on the window, then if this is run with 
            # create_plot == 'y' or 'Y', then the current plot on the window will double
            if create_plot == 'y':
                self.canvas_options = FigureCanvasTkAgg(figure_House, Frame_house)
                self.canvas_options.get_tk_widget().pack(fill = BOTH, expand = 1) 

    root = Tk()
    # Adjusts window size; width x height + x + y (x,y coordinates for where window will be placed on screen)
    root.geometry("1470x720+30+20")
    app = Application(master=root)
    # Calls and opens GUI
    app.mainloop()