# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 11:27:36 2016

@author: akibdpt
"""
from .multiplyForm import multiplyForm


"""
 * A discrete-form specifies a form with a discrete operational
 * behavior, corresponding a mathematical set: any two individuals in the
 * form are either identical or disjoint. A discrete form is maximal
 * if no two individuals are identical, and the individuals' attribute forms,
 * if any, are also maximal.
 * 
 * The DiscreteForm class extends on the {MultiplyForm} class and
 * specifies the discrete operational behavior of forms.
"""
class discrete(multiplyForm):
    
    def __init__(self, sort):
        super().__init__(sort)
        self.type = 'form-discrete'
    
    """
    print out the form for easier assessment
    structure:
    |sort name: n elements
     _ individual 1
     _ individual 2
         ...
     _ individual n
      |attribute1 sort name: m elements
       _ attribute1 1
      |attribute2 sort name: p elements
       _ attribute1 2
    """
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' _ ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
    """
    inserting an element into the form
    """
    def insert(self, ind):
        if not self.sort == ind.sort:
            raise ValueError('not of the same sort')
            return
        thisInd = self.first()
        index = 0
        while thisInd is not None:
            if thisInd.compareValue(ind) == 0:
                if thisInd.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            elif thisInd.compareValue(ind) == 1:
                self.add(index, ind)
                return
            else: 
                thisInd = self.next(thisInd)      
                index = index+1
        self.add(index, ind)
        return
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    def maximalize(self):
        if self.maximal:
            return
    
        self.maximal = True
        
        if len(self.elements) == 0:
            return
        
        ind = self.first()
        nextInd = self.next(ind)
        disorganized = []
        while nextInd is not None:
            #if nextInd is Nil, remove it and assign nextInd as the next one
            if nextInd.isNil():
                self.remove(nextInd)
                ind = nextInd
                nextInd = self.next(nextInd)
            #if ind is Nil(should only happen for the first element), move 
            #   ind and nextInd one step forward
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.next(nextInd)
            #if ind is equals to nextInd, sum up their attributes and remove 
            #   nextInd, assign nextInd one step forward 
            elif ind.compareValue(nextInd) == 0:
                if ind.hasAttribute():
                    ind.attribute.sum(nextInd.attribute)                
                toBeRemoved = nextInd                
                nextInd = self.next(nextInd)
                self.remove(toBeRemoved)
            #if ind is greater than nextInd, remove nextInd, add it to 
            #   disorganized to be added back in later in the correct place, 
            #   assign nextInd one step forward
            elif ind.compareValue(nextInd) == 1:
                toBeRemoved = nextInd                
                nextInd = self.next(nextInd)
                disorganized.append(toBeRemoved)
                self.remove(toBeRemoved)
            #if ind is lesser than nextInd, nothing happens, more ind and 
            #   nextInd one step forward
            else:
                ind = nextInd
                nextInd = self.next(nextInd)     
        
        #add all the disorganized element back in the form again in the correct place
        for ins in disorganized:
            self.insert(ins)
            
        #maximalize the attributes of the individuals in the form
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()                
        return
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
        
        thisInd = self.first()
        otherInd = other.first()
        index = 0
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, sum
                if thisInd.hasAttribute():
                    thisInd.attribute.sum(otherInd.attribute)
                thisInd = self.next(thisInd)                            
                index = index+1
                otherInd = other.next(otherInd)
            elif thisInd.compareValue(otherInd) == 1:
                self.add(index, otherInd)                            
                index = index+1
                otherInd = other.next(otherInd)
            else:                          
                index = index+1
                thisInd =self.next(thisInd)       
        while otherInd is not None:
            self.add(index, otherInd)                            
            index = index+1
            otherInd = other.next(otherInd)
        return

    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
            
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, calculate difference
                if thisInd.hasAttribute():
                    thisInd.attribute.difference(otherInd.attribute)
                if not thisInd.hasAttribute():
                    toBeRemovedInd = thisInd
                    thisInd = self.next(thisInd)
                    self.remove(toBeRemovedInd)
                otherInd = other.next(otherInd)
            elif thisInd.compareValue(otherInd) == 1:                            
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)       
        return

    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, product
                if thisInd.hasAttribute():
                    thisInd.attribute.product(otherInd.attribute)
                """
                if not thisInd.hasAttribute() and otherInd.hasAttribute():
                    toBeRemovedInd = thisInd
                    thisInd = self.next(thisInd)
                    otherInd = other.next(otherInd)
                    self.remove(toBeRemovedInd)
                else:
                    thisInd = self.next(thisInd)                            
                    otherInd = other.next(otherInd)
                """
                thisInd = self.next(thisInd)                            
                otherInd = other.next(otherInd)
            elif thisInd.compareValue(otherInd) == 1:                
                otherInd = other.next(otherInd)
            else:                             
                toBeRemovedInd = thisInd
                thisInd = self.next(thisInd)
                self.remove(toBeRemovedInd)
        while thisInd is not None:   
            toBeRemovedInd = thisInd
            thisInd = self.next(thisInd)
            self.remove(toBeRemovedInd)
        return

    """
    symdifference of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.symdifference(formB)
      -->formA: a,       d
    """
    def symdifference(self,other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        otherInd = other.first()
        index = 0
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, calculate difference
                if thisInd.hasAttribute():
                    thisInd.attribute.symdifference(otherInd.attribute)
                if not thisInd.hasAttribute():
                    toBeRemovedInd = thisInd
                    thisInd = self.next(thisInd)
                    self.remove(toBeRemovedInd)
                otherInd = other.next(otherInd)
                
            elif thisInd.compareValue(otherInd) == 1:
                self.add(index, otherInd)                            
                index = index+1
                otherInd = other.next(otherInd)
            else:                             
                index = index+1
                thisInd =self.next(thisInd)       
        while otherInd is not None:
            self.add(index, otherInd)                            
            index = index+1
            otherInd = other.next(otherInd)
        return 

    """
    partition of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
         
        _common = formA.partition(formB)
      -->formA: a
         formB:          d
         common:   b, c
    """
    def partition(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        otherInd = other.first()
        common = discrete(self.sort)
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:                

                attForm = otherInd.duplicateInd()
                
                if attForm.hasAttribute():
                    attForm.attribute.product(thisInd.attribute)
                
                if not (not attForm.hasAttribute() and (thisInd.hasAttribute() or otherInd.hasAttribute())):
                    common.insert(attForm)    
                    
                    if thisInd.hasAttribute():
                        thisInd.attribute.difference(attForm.attribute)
                    if otherInd.hasAttribute():
                        otherInd.attribute.difference(attForm.attribute)
                    
                    if not thisInd.hasAttribute():
                        toBeRemovedInd = thisInd
                        thisInd = self.next(thisInd)
                        self.remove(toBeRemovedInd)
                    else:
                        thisInd = self.next(thisInd)
                        
                    if not otherInd.hasAttribute():
                        toBeRemovedInd = otherInd
                        otherInd = other.next(otherInd)
                        other.remove(toBeRemovedInd)
                    else:
                        otherInd = other.next(otherInd)
                else:
                    thisInd = self.next(thisInd)
                    otherInd = other.next(otherInd)                
                    
            elif thisInd.compareValue(otherInd) == 1:      
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)       
        return common
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                    return False
                thisInd = self.next(thisInd)
                otherInd = other.next(otherInd)
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.next(otherInd)
            else:                             
                return False
        return thisInd is None