# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 14:19:32 2016

@author: akibdpt
"""

from .sort import sort
from ..form.metaForm import metaForm

"""
 * A Disjunctive Sort specifies a disjunctive composition of
 * {Sort}s under the sum operation. Under the sum operation,
 * a form of the resulting sort may contain a form of every operand sort.
 * If any component sort is an unnamed disjunctive sort, instead,
 * its components become part of the disjunctive composition.
 * If any component sort is part of another, unnamed component sort, the former
 * component sort is not included in the composition.
 * A disjunctive sort may have a name assigned.
 * 
 * The DisjunctiveSort class represents a disjunctive sort additionally
 * as an ordered set of component sorts. 
 * In the definition of a composite sort, the operation of sum is denoted by '+'.
 * In its canonical version, component sorts cannot themselves be disjunctive sorts
 * and XXX.
 * (a + b) + (a + c) -> a + b + c.
"""

class disjunctiveSort(sort):
    separator = ' + '
    
    """
    initialize the sort, a disjunctive sort should not contain any other disjunctive sort.
    needs to go through all the component and compile all the subordinate sorts
    e.g. (a + b) + (a + c) -> a + b + c
    then make the description for the sort
    """
    def __init__(self, name, component = []):
        if self.checkSort(name) == False:
            pass
            raise ValueError('sort name already exists')
        super().__init__(name)
        self.type = 'sort-disjunctive'
        
        #compiling all sorts from the component
        uniqueComp = []
        for i in component:
            if i.type == 'sort-disjunctive':
                for j in i.component:
                    if j not in uniqueComp:
                        uniqueComp.append(j)
            elif i not in uniqueComp:
                uniqueComp.append(i)
                
        #combining attribute sorts with the same base sort
        iIndex = 0
        while iIndex < len(uniqueComp):
            i = uniqueComp[iIndex]
            jIndex = iIndex + 1
            while jIndex < len(uniqueComp):
                j = uniqueComp[jIndex]
                if i.type == 'sort-attribute' and j.type == 'sort-attribute' and i.base.name == j.base.name:
                    nameList = []
                    if i.weight.type == 'sort-disjunctive':
                        for comp in i.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(i.weight.name)
                        
                    if j.weight.type == 'sort-disjunctive':
                        for comp in j.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(j.weight.name)
                        
                    
                    newDisj = disjunctiveSort(disjunctiveSort.separator.join(nameList), [i.weight, j.weight])
                    i.weight = newDisj
                    i.name= i.base.name+'_'+'('+i.weight.name.replace(" ", "")+')'
                    uniqueComp.remove(j)
                else:
                    jIndex += 1
            iIndex += 1
            
        #sort the list
        uniqueComp.sort() 
        
        #assign to self.component
        self.component = uniqueComp
        
        self.description = self.description(uniqueComp)
        return
        
    def __eq__(self,other):
        for i in self.component:
            check = False
            for j in other.component:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
    
    """
    add a component sort to this sort
    """
    def addComponent(self, sort):
        for i in self.component:
            if sort.name == i.name:
                return
        self.component.append(sort)
    
    """
    creating the description for this sort
    """
    def description(self, component):
        if len(component) <= 1:
            desc = ''
            for i in component:
                desc += i.description
        else:
            desc = '('
            for i in component:
                if desc == '(':
                    desc += i.description
                else:
                    desc += disjunctiveSort.separator + i.description
            desc += ')'
        return desc
        
    """
    checking if this sort contain other sort
    """
    def contains(self, other):
        if other.type == 'sort-disjunctive':
            for i in other.component:
                Check = False
                for j in self.component:
                    if j.contains(i):
                        Check = True
                if Check == False:
                    return False
            return True
        else:
            for i in self.component:
                if i.contains(other):
                    return True
        return False
        
    """
    dimension of this sort is 0
    """
    def dimension(self):
        return 0
        
    """
    creating a new metaform of this sort
    """    
    def newForm(self):
        try:
            form = metaForm(self)
            return form
        except:
            print('exception')
    
    