# -*- coding: utf-8 -*-
"""
Created on Mon Aug  8 15:44:26 2016

@author: akibdpt
"""
"""
 * A Primitive Sort is a {SimpleSort} defined by
 * a characteristic individual (see also {sortal.ind.individual})
 * and a behavioral category that specifies the operational behavior of
 * its {sortal.form.Form}s, that is, collections of individuals.
 * A primitive sort also always has a name assigned.
 * 
 * The PrimitiveSort class represents a primitive sort additionally by
 * its characteristic individual class, behavior class, corresponding arguments,
 * default {sortal.map.Mapping}, and dimensionality.
 * The canonical version of a primitive sort is the sort itself, unless
 * the sort has been created as a duplicate of another primitive sort,
 * in which case it takes its canonical version from that other sort.
 * It implements the {SimpleSort} and {sortal.io.Parsing}
 * interfaces.
"""
from .simpleSort import simpleSort
from .parameter.argument import argument

class primitiveSort(simpleSort):
    categories = {}
    parameter = {}
    defaultMapping = {}
    dimensionality = {}
    
    def __init__(self, name, characteristic, arg = None):
        if self.checkSort(name) == False:
            pass
            raise ValueError('sort name already exists')
        super().__init__(name)
        self.type = 'sort-primitive'
        self.characteristic = characteristic
        self.behavior = primitiveSort.categories[characteristic]
        self.parameter = primitiveSort.parameter[characteristic]
        self.description = str(characteristic)

        if self.parameter != argument.argumentType.NONE:
            self.arguments = argument.defineArgument(self.parameter,arg)
            if self.arguments is not None:
                self.description += str(self.arguments.argValue())
        
    """
    registering each individual with its category, parameter, defaultMapping and dimensionality
    e.g. a point is of discrete category, no parameter, similarity mapping and discrete dimensionality
         a line segment is of interval category, no parameter, similarity mapping and linear dimensionality
         a label is of discrete category, no parameter, exact mapping and discrete dimensionality
         a weight is of ordinal category, 'upperbound' parameter, exact mapping and linear dimensionality
         etc...
    this function is called specifically in sortal/__init__.py and register all available individuals for
    rule analysis and detection
    """
    def register(characteristic, b, p, m, d):
        primitiveSort.categories[characteristic] = b
        primitiveSort.parameter[characteristic] = p
        primitiveSort.defaultMapping[characteristic] = m
        primitiveSort.dimensionality[characteristic] = d
        
    def base(self):
        return self
            
    """
    get respective behavior(category) from an individual
    """
    def getBehavior(characteristic):
        return primitiveSort.categories[characteristic]
    """
    get respective parameter from an individual
    """
    def getParameter(self):
        return primitiveSort.parameter[self.characteristic]
    """
    get respective mapping from an individual
    """
    def mapping(self):
        return primitiveSort.defaultMapping[self.characteristic]
    """
    get respective dimensionality from an individual
    """
    def dimension(self):
        return primitiveSort.dimensionality[self.characteristic]
        
    """
    checking if this sort contains another sort i.e. the two sorts are the same
    """
    def contains(self, other):
        if other.type == 'sort-primitive' and other.name == self.name:
            return True
        return False
        
    """
    creating a new form (discrete form / ordinal form / interval form) of primitive sort
    """    
    def newForm(self):
        try:
            form = self.behavior(self)
            return form
        except:
            print('exception')

    
