# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:40:57 2016

@author: akibdpt
"""
from ..struct.coordinate import coordinate
from ..struct.vectorFloat import vector2Dfloat, vector3Dfloat
from ..struct.vector import vector3D
from sympy import sin, cos, N

from mpmath import mp, sin as sinF, cos as cosF, sign, chop
#from decimal import Decimal


mp.dps = vector2Dfloat.mpdps

c0 = coordinate(0)
c1 = coordinate(1)

class affineTransform2D():
    SIZE = 3 #4
    
    
    
    def __init__(self, reflection = None):
        self.type = 'transform-affineTransform2D'
        self.matrix = [[1,0,0],
                       [0,1,0]]
        if reflection is not None:
            self.matrix[reflection][reflection] = -1
        self.index = 0
        return
        
    def __str__(self):
        return str(self.matrix[0])+'\n'+str(self.matrix[1])+'\n'
        
    def printTransf(self):
        print('matrix:')
        for i in self.matrix:
            #print('|',i[0],i[1],i[2])
            print('|',round(i[0],3),round(i[1],3),round(i[2],3))
        
    
    def scalar(self, n, m):
        if n < 0 or n >= affineTransform2D.SIZE or m < 0 or m >= affineTransform2D.SIZE:
            raise IndexError
        return self.matrix[n][m]
    
    def convertFloat(self):
        result = affineTransform2D()
        for i in range(affineTransform2D.SIZE):
            for j in range(affineTransform2D.SIZE):
                if isinstance(self.matrix[i][j], coordinate): 
                    result.matrix[i][j] = self.matrix[i][j].value
                elif isinstance(self.matrix[i][j], float):
                    raise ValueError('self.matrix[i][j] should not be of float value since it is highly inaccurate')
                else:
                    result.matrix[i][j] = mp.mpf(self.matrix[i][j])
        return result
        
    def equalsFloat(self, other):
        for i in range(affineTransform2D.SIZE):
            for j in range(affineTransform2D.SIZE):
                if round(self.matrix[i][j], vector2Dfloat.p) != round(other.matrix[i][j], vector2Dfloat.p):
                    print(round(self.matrix[i][j], vector2Dfloat.p), round(other.matrix[i][j], vector2Dfloat.p))
                    return False
        return True
        


    @staticmethod
    def translation(vector):
        result = affineTransform2D()
        result.matrix = [[c1, c0, vector.x],
                         [c0, c1, vector.y]]
        return result
        
    @staticmethod
    def translationFloat(vector):
        result = affineTransform2D()
        result.matrix = [[1, 0, vector.x],
                         [0, 1, vector.y]]
        return result
        
    @staticmethod
    def scale(factor):
        result = affineTransform2D()
        result.matrix = [[coordinate(factor), c0                , c0],
                         [c0                , coordinate(factor), c0]]
        return result
        
    @staticmethod
    def scaleFloat(factor):
        result = affineTransform2D()
        result.matrix = [[factor, 0     , 0],
                         [0     , factor, 0]]
        return result
        
    @staticmethod
    def rotation(theta):
        result = affineTransform2D()
        result.matrix = [[coordinate(cos(theta)), coordinate(-sin(theta)), c0],
                         [coordinate(sin(theta)),  coordinate(cos(theta)), c0]]
        return result
        
    @staticmethod
    def rotationFloat(theta):
        result = affineTransform2D()
        result.matrix = [[(cosF(theta)), (-sinF(theta)), 0],
                         [(sinF(theta)), ( cosF(theta)), 0]]
        return result
        
    @staticmethod
    def scaledRotation(theta, lengthFrom, lengthTo):
        result = affineTransform2D()
        l = lengthTo / lengthFrom
        result.matrix = [[(coordinate(cos(theta))*l), (coordinate(-sin(theta))*l), c0],
                         [(coordinate(sin(theta))*l),  (coordinate(cos(theta))*l), c0]]
        return result
        
    @staticmethod
    def scaledRotationFloat(theta, lengthFrom, lengthTo):
        result = affineTransform2D()
        l = lengthTo / lengthFrom
        result.matrix = [[cos(theta)*l, -sin(theta)*l, 0],
                         [sin(theta)*l,  cos(theta)*l, 0]]
        return result
        
    @staticmethod
    def reflectionX():
        result = affineTransform2D()
        result.matrix = [[1, 0,  0],
                         [0, -1, 0]]
        return result
        
    @staticmethod
    def reflectionXY():
        result = affineTransform2D()
        result.matrix = [[0, 1, 0],
                         [1, 0, 0]]
        return result
                
    def combine(self, other):
        result = affineTransform2D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform2D.SIZE-1):
            for j in range(affineTransform2D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j]).simplify()
        result.matrix[0][2] = a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2].simplify()
        result.matrix[1][2] = a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2].simplify()
        return result
    
    def combineFloat(self, other):
        result = affineTransform2D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform2D.SIZE-1):
            for j in range(affineTransform2D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j])
        result.matrix[0][2] = a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]
        result.matrix[1][2] = a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]
        return result
        
    @staticmethod    
    def similarityTransform(fromE, toE, exact, index):
        dirFrom = fromE[1].subtract(fromE[0])
        dirTo = toE[1].subtract(toE[0])
        
        fromLength = dirFrom.length()
        toLength = dirTo.length()

        transformation= None
        transformationRefl = None
        if exact: # == False:
            if index != 2:
                baseRotation = affineTransform2D.scaledRotation(dirFrom.angle(dirTo), fromLength, toLength)
                transformation = affineTransform2D.translation(toE[0].subtract(fromE[0].transform(baseRotation))).combine(baseRotation)
                transformation.index = 1
            if index != 1:
                reflRotation = affineTransform2D.scaledRotation(dirFrom.transform(affineTransform2D.reflectionX()).angle(dirTo), fromLength, toLength).combine(affineTransform2D.reflectionX())
                transformationRefl = affineTransform2D.translation(toE[0].subtract(fromE[0].transform(reflRotation))).combine(reflRotation)
                transformationRefl.index = 2
                
        else:
            if index != 2:
                baseRotation = affineTransform2D.scaledRotationFloat(dirFrom.angle(dirTo), fromLength, toLength)
                transformation = affineTransform2D.combineFloat(affineTransform2D.translationFloat(toE[0].subtract(fromE[0].transform(baseRotation))),baseRotation)
                transformation.index = 1
            if index != 1:
                reflRotation = affineTransform2D.scaledRotationFloat(dirFrom.transform(affineTransform2D.reflectionX()).angle(dirTo), fromLength, toLength).combineFloat(affineTransform2D.reflectionX())
                transformationRefl = affineTransform2D.translationFloat(toE[0].subtract(fromE[0].transform(reflRotation))).combineFloat(reflRotation)
                transformationRefl.index = 2
        result = [transformation, transformationRefl]
        return result


class affineTransform3D():
    SIZE = 4
    
    rotX = 0
    rotY = 1
    rotZ = 2
    
    def __init__(self, reflection = None):
        self.type = 'transform-affineTransform3D'
        self.matrix = [[1,0,0,0],
                       [0,1,0,0],
                       [0,0,1,0]]
        if reflection is not None:
            self.matrix[reflection][reflection] = -1
        self.index = 0
        return
        
    def __str__(self):
        return str(self.matrix[0])+'\n'+str(self.matrix[1])+'\n'+str(self.matrix[2])+'\n'
        
    def printTransf(self):
        for i in self.matrix:
            try:
                print('|\t',round(i[0],3),'\t',round(i[1],3),'\t',round(i[2],3),'\t',round(i[3],3))
            except TypeError:
                print('|',i[0],'\t',i[1],'\t',i[2],'\t',i[3])
        print('')
        
    
    def scalar(self, n, m):
        if n < 0 or n >= affineTransform3D.SIZE or m < 0 or m >= affineTransform3D.SIZE:
            raise IndexError
        return self.matrix[n][m]
    
    def convertFloat(self):
        result = affineTransform3D()
        for i in range(affineTransform3D.SIZE):
            for j in range(affineTransform3D.SIZE):
                if isinstance(self.matrix[i][j], coordinate): 
                    result.matrix[i][j] = self.matrix[i][j].value
                elif isinstance(self.matrix[i][j], float):
                    raise ValueError('self.matrix[i][j] should not be of float value since it is highly inaccurate')
                else:
                    result.matrix[i][j] = mp.mpf(self.matrix[i][j])
        return result
        
    def equalsFloat(self, other):
        for i in range(affineTransform3D.SIZE):
            for j in range(affineTransform3D.SIZE):
                if round(self.matrix[i][j], vector3Dfloat.p) != round(other.matrix[i][j], vector3Dfloat.p):
                    print(round(self.matrix[i][j], vector3Dfloat.p), round(other.matrix[i][j], vector3Dfloat.p))
                    return False
        return True


    @staticmethod
    def tensorProduct(vector)  :     
        result = affineTransform3D()
        coords = [vector.x,vector.y,vector.z]
        for i in range(len(result.matrix)):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j] = coords[i] * coords[j]
        return result;

    @staticmethod
    def crossProduct(vector):
        result = affineTransform3D()
        result.matrix = [[0       , -vector.z, vector.y , 0],
                         [vector.z , 0       , -vector.x, 0],
                         [-vector.y, vector.x , 0       , 0]]
        return result
        
    def add(self, other):
        result = affineTransform3D()
        for i in range(len(result.matrix)):
            for j in range(len(result.matrix[i])):
                result.matrix[i][j] = self.matrix[i][j]+other.matrix[i][j]
        return result
        
    def tScale(self,factor):
        result = affineTransform3D()
        for i in range(len(result.matrix)):
            for j in range(affineTransform3D.SIZE):
                result.matrix[i][j] = factor * self.matrix[i][j]
        return result
        
    def transpose(self):
        result = affineTransform3D()
        for i in range(affineTransform3D.SIZE-1):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j] = self.matrix[j][i]
        if isinstance(self.matrix[0][0], coordinate):
            result.matrix[0][3] = c0
            result.matrix[1][3] = c0
            result.matrix[2][3] = c0
        else:
            result.matrix[0][3] = 0
            result.matrix[1][3] = 0
            result.matrix[2][3] = 0
        return result



    @staticmethod
    def translation(vector):
        result = affineTransform3D()
        result.matrix = [[c1, c0, c0, vector.x],
                         [c0, c1, c0, vector.y],
                         [c0, c0, c1, vector.z]]
        return result
        
    @staticmethod
    def translationFloat(vector):
        result = affineTransform3D()
        result.matrix = [[1, 0, 0, vector.x],
                         [0, 1, 0, vector.y],
                         [0, 0, 1, vector.z]]
        return result
        
    @staticmethod
    def scale(factor):
        result = affineTransform3D()
        result.matrix = [[coordinate(factor), c0, c0, c0],
                         [c0, coordinate(factor), c0, c0],
                         [c0, c0, coordinate(factor), c0]]
        return result
        
    @staticmethod
    def scaleFloat(factor):
        result = affineTransform3D()
        result.matrix = [[factor, 0     , 0     , 0],
                         [0     , factor, 0     , 0],
                         [0     , 0     , factor, 0]]
        return result
    
    
    @staticmethod
    def rotation(vector, sinSq, cos):
        squareLength = vector.dotproduct(vector)
        result = affineTransform3D().tScale(cos)
        tf = affineTransform3D.tensorProduct(vector).tScale(c1 / squareLength).tScale(c1-cos)
        result = result.add(tf)
        tf = affineTransform3D.crossProduct(vector).tScale((sinSq / squareLength) ** coordinate(0.5))
        result = result.add(tf)
        return result
        
    @staticmethod
    def rotationFloat(vector, sinSq, cos):
        
        #print('vector:',vector)
        #print('sinsq', sinSq, 'cos', cos)
        squareLength = vector.dotproduct(vector)
        result = affineTransform3D().tScale(cos)
        #result.printTransf()
        tf = affineTransform3D.tensorProduct(vector).tScale(mp.mpf('1')/squareLength).tScale(mp.mpf('1')-cos)
        result = result.add(tf)
        #tf.printTransf()
        tf = affineTransform3D.crossProduct(vector).tScale((sinSq/squareLength) ** mp.mpf('0.5'))
        #tf.printTransf()
        result = result.add(tf)
        return result
        
    @staticmethod
    def scaledRotation(vector, sinSq, cosSq, cosSign, fSq):
        squareLength = vector.dotproduct(vector)
        fcos = (cosSq * fSq) ** coordinate(0.5)
        if cosSign < 0:
            fcos = -fcos
        result = affineTransform3D().tScale(fcos)
        tf = affineTransform3D.tensorProduct(vector).tScale(c1/squareLength).tScale(fSq**coordinate(0.5)-fcos)
        result = result.add(tf)
        fsinSq = fSq * (sinSq/squareLength)
        tf = affineTransform3D.crossProduct(vector).tScale(fsinSq**coordinate(0.5))
        result = result.add(tf)
        return result
        
    @staticmethod
    def scaledRotationFloat(vector, sinSq, cosSq, cosSign, fSq):
        #print('vector:',vector)
        #print('sinSq:',sinSq,'cosSq:', cosSq,'cosSign:', cosSign,'fSq:', fSq)
        squareLength = vector.dotproduct(vector)
        fcos = (cosSq * fSq)**mp.mpf('0.5')
        if cosSign < 0:
            fcos = -fcos
        result = affineTransform3D().tScale(fcos)
        tf = affineTransform3D.tensorProduct(vector).tScale(mp.mpf('1')/squareLength).tScale(fSq**mp.mpf('0.5')-fcos)
        result = result.add(tf)
        fsinSq = fSq * (sinSq/squareLength)
        tf = affineTransform3D.crossProduct(vector).tScale(fsinSq**mp.mpf('0.5'))
        result = result.add(tf)
        return result
                        
    def combine(self, other):
        result = affineTransform3D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform3D.SIZE-1):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j]+a[i][2]*b[2][j]).simplify()
        result.matrix[0][3] = (a[0][0]*b[0][3]+a[0][1]*b[1][3]+a[0][2]*b[2][3]+a[0][3]).simplify()
        result.matrix[1][3] = (a[1][0]*b[0][3]+a[1][1]*b[1][3]+a[1][2]*b[2][3]+a[1][3]).simplify()
        result.matrix[2][3] = (a[2][0]*b[0][3]+a[2][1]*b[1][3]+a[2][2]*b[2][3]+a[2][3]).simplify()
        return result
    
    def combineFloat(self, other):
        result = affineTransform3D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform3D.SIZE-1):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j]+a[i][2]*b[2][j])
        result.matrix[0][3] = (a[0][0]*b[0][3]+a[0][1]*b[1][3]+a[0][2]*b[2][3]+a[0][3])
        result.matrix[1][3] = (a[1][0]*b[0][3]+a[1][1]*b[1][3]+a[1][2]*b[2][3]+a[1][3])
        result.matrix[2][3] = (a[2][0]*b[0][3]+a[2][1]*b[1][3]+a[2][2]*b[2][3]+a[2][3])
        return result
        
    def floatChop(self):
        for i in self.matrix:
            for j in i:
                j = chop(j)
                
    @staticmethod
    def reflectionX():
        result = affineTransform3D()
        result.matrix = [[1,  0,  0, 0],
                         [0, -1,  0, 0],
                         [0,  0, -1, 0]]
        return result
        
    @staticmethod    
    def similarityTransform(fromE, toE, exact, index):   
        result = []
        uf = fromE[1].subtract(fromE[0])
        vf = fromE[2].subtract(fromE[0])
        ut = toE[1].subtract(toE[0])
        vt = toE[2].subtract(toE[0])
        """
        luf = uf.length()
        lvf = vf.length()
        lut = ut.length()
        lvt = vt.length()
        
        for i in fromE:
            print('{{{{', i)
        for i in toE:
            print('}}}}', i)
        print('.|.',uf.dotproduct(vf) / (luf * lvf),  ut.dotproduct(vt) / (lut * lvt),luf / lvf, lut / lvt)
        if (uf.dotproduct(vf) / (luf * lvf) != ut.dotproduct(vt) / (lut * lvt)
                or luf / lvf != lut / lvt ):
            print('why...')
            return result
        """
        
        if exact:

            #normal vectors of from set and to set
            normalfrom = uf.product(vf)
            normalto = ut.product(vt)
            
            #mapping the normal vector of the from set and to set together
            step1 = affineTransform3D()
            axis = normalfrom.product(normalto)
            if not axis.isZero():
                sqLengths = normalfrom.dotproduct(normalfrom) * normalto.dotproduct(normalto)
                sinSq = axis.dotproduct(axis) / sqLengths
                cos = normalfrom.dotproduct(normalto) / sqLengths.sqrt()
                step1 = affineTransform3D.rotation(axis, sinSq, cos)
                
            #transforming the vectors of from set, then find the direction vectors of the 2 sets
            directionfrom = fromE[2].transform(step1).subtract(fromE[1].transform(step1))
            directionto = toE[2].subtract(toE[1])
            
            #finding the scaled rotation to map directionfrom to directionto (no translation)
            axis = directionfrom.product(directionto)
            sqLengths = directionfrom.dotproduct(directionfrom) * directionto.dotproduct(directionto)
            cosSq = directionfrom.dotproduct(directionto) ** 2 / sqLengths
            scalarSq = directionto.dotproduct(directionto) / directionfrom.dotproduct(directionfrom)
        
            if not axis.isZero():
                sinSq = axis.dotproduct(axis) / sqLengths
                step2 = affineTransform3D.scaledRotation(axis, sinSq, cosSq, directionfrom.dotproduct(directionto).sign(), scalarSq)
            else:
                step2 = affineTransform3D.scaledRotation(normalto, c0, cosSq, directionfrom.dotproduct(directionto).sign(), scalarSq)
            
            #180 degree rotation matrix
            rot180 = affineTransform3D.rotation(directionto, c0, coordinate(-1))
            
            axis = normalto.product(vector3D(0,0,1))
            rotZ = affineTransform3D()
            if not axis.isZero():
                sqLengths = normalto.dotproduct(normalto)
                sinSq = axis.dotproduct(axis) / sqLengths
                cos = normalto.z / (sqLengths ** coordinate(0.5))
                rotZ = affineTransform3D.rotation(axis, sinSq, cos)
                
            baseRots = step2.combine(step1)
            tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(baseRots))).combine(baseRots)

            transformation= None
            transformationRefl = None
            
            """
            print('base rotation:')
            baseRots.printTransf()
            print('tf:')
            tf.printTransf()
            print('.:.__',fromE[0].transform(tf), toE[0])
            print('.:.__',fromE[1].transform(tf), toE[1])
            print('.:.__',fromE[2].transform(tf), toE[2])
            """

            if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
                if index != 2:
                    transformation = tf
                    transformation.index = 1
                if index != 1:
                    reflRots = baseRots.combine(rotZ.transpose()).combine(affineTransform3D(2)).combine(rotZ)
                    transformationRefl = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(reflRots))).combine(reflRots)
                    transformationRefl.index = 2
                result = [transformation,transformationRefl]
                return result

            tf = rot180.combine(baseRots)
            tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(tf))).combine(tf)
            
            """
            print('\n.:.__',fromE[0].transform(tf), toE[0])
            print('.:.__',fromE[1].transform(tf), toE[1])
            print('.:.__',fromE[2].transform(tf), toE[2])
            """
            
            if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
                if index != 2:
                    transformation = tf
                    transformation.index = 1
                if index != 1:
                    reflRots = baseRots.combine(rotZ.transpose()).combine(affineTransform3D(2)).combine(rotZ)
                    tf = rot180.combine(reflRots)
                    transformationRefl = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(tf))).combine(tf)
                    transformationRefl.index = 2
                result = [transformation,transformationRefl]
                return result

        else:
            #normal vectors of from set and to set
            normalfrom = uf.product(vf)
            normalto = ut.product(vt)
            
            #mapping the normal vector of the from set and to set together
            step1 = affineTransform3D()
            axis = normalfrom.product(normalto)
            """
            print('\n~~~~~~~~')
            print('from:',fromE[0],fromE[1],fromE[2])
            print('to:  ',toE[0],toE[1],toE[2])
            print('\nnormalfrom:',normalfrom,'normalto:', normalto)
            print('axis:', axis)
            """
            if not axis.isZero():
                sqLengths = normalfrom.dotproduct(normalfrom) * normalto.dotproduct(normalto)
                sinSq = axis.dotproduct(axis) / sqLengths
                cos = normalfrom.dotproduct(normalto) / sqLengths.sqrt()
                step1 = affineTransform3D.rotationFloat(axis, sinSq, cos)
            """
            print('step 1: map normal')
            step1.printTransf()            
            """
            
            #transforming the vectors of from set, then find the direction vectors of the 2 sets
            directionfrom = fromE[2].transform(step1).subtract(fromE[1].transform(step1))
            directionto = toE[2].subtract(toE[1])
            """
            print('from1:', fromE[1].transform(step1),'from2:',fromE[2].transform(step1))
            print('to1:', toE[1],'to2:',toE[2])
            print('dirFrom:', directionfrom,'dirTo:',directionto)
            """
            #finding the scaled rotation to map directionfrom to directionto (no translation)
            axis = directionfrom.product(directionto)
            sqLengths = directionfrom.dotproduct(directionfrom) * directionto.dotproduct(directionto)
            cosSq = directionfrom.dotproduct(directionto) ** 2 / sqLengths
            scalarSq = directionto.dotproduct(directionto) / directionfrom.dotproduct(directionfrom)
            if not axis.isZero():
                sinSq = axis.dotproduct(axis) / sqLengths
                step2 = affineTransform3D.scaledRotationFloat(axis, sinSq, cosSq,sign(directionfrom.dotproduct(directionto)), scalarSq)
            else:
                step2 = affineTransform3D.scaledRotationFloat(normalto, 0, cosSq, sign(directionfrom.dotproduct(directionto)), scalarSq)
            """
            print('step 2: map direction')
            step2.printTransf()            
            """
            #180 degree rotation matrix
            rot180 = affineTransform3D.rotationFloat(directionto, 0, mp.mpf('-1'))
            
            axis = normalto.product(vector3Dfloat(0,0,1))
            rotZ = affineTransform3D()
            if not axis.isZero():
                sqLengths = normalto.dotproduct(normalto)
                sinSq = axis.dotproduct(axis) / sqLengths
                cos = normalto.z / (sqLengths ** mp.mpf('0.5'))
                rotZ = affineTransform3D.rotationFloat(axis, sinSq, cos)
            
            baseRots = step2.combineFloat(step1)
            tf = affineTransform3D.translationFloat(toE[0].subtract(fromE[0].transform(baseRots))).combineFloat(baseRots)
            transformation= None
            transformationRefl = None

            """
            print('base rotation:')
            baseRots.printTransf()
            print('tf:')
            tf.printTransf()
            print('.:.__',fromE[0].transform(tf), toE[0])
            print('.:.__',fromE[1].transform(tf), toE[1])
            print('.:.__',fromE[2].transform(tf), toE[2])
            """
            
            if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
                if index != 2:
                    transformation = tf
                    transformation.index = 1
                if index != 1:
                    reflRots = rotZ.transpose().combineFloat(affineTransform3D(2)).combineFloat(rotZ).combineFloat(baseRots)
                    transformationRefl = affineTransform3D.translationFloat(toE[0].subtract(fromE[0].transform(reflRots))).combineFloat(reflRots)
                    transformationRefl.index = 2
                result = [transformation,transformationRefl]
                return result

            tf = rot180.combineFloat(baseRots)
            tf = affineTransform3D.translationFloat(toE[0].subtract(fromE[0].transform(tf))).combineFloat(tf)
            
            """
            print('\n.:.__',fromE[0].transform(tf), toE[0])
            print('.:.__',fromE[1].transform(tf), toE[1])
            print('.:.__',fromE[2].transform(tf), toE[2])
            """
            
            if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
                if index != 2:
                    transformation = tf
                    transformation.index = 1
                if index != 1:
                    reflRots = rotZ.transpose().combineFloat(affineTransform3D(2)).combineFloat(rotZ).combineFloat(baseRots)
                    tf = rot180.combineFloat(reflRots)
                    transformationRefl = affineTransform3D.translationFloat(toE[0].subtract(fromE[0].transform(tf))).combineFloat(tf)
                    transformationRefl.index = 2
                result = [transformation,transformationRefl]
                return result
        raise ValueError('no appropriate transformation found')

